package dev.pfaff.altargui.glfw;

import dev.pfaff.altargui.host.ContextForker;
import dev.pfaff.altargui.host.ForkedContext;
import dev.pfaff.altargui.host.Host;
import dev.pfaff.altargui.host.IWindow;
import dev.pfaff.altargui.util.BasicFuture;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.altargui.util.listenable.ConstantValue;
import dev.pfaff.altargui.util.listenable.Value;
import dev.pfaff.dyl4j.BundledLibraryLinker;
import dev.pfaff.glass.Arch;
import dev.pfaff.glass.OS;
import dev.pfaff.glass.glfw.GLFWwindow;
import dev.pfaff.glass.opengl.GL;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.io.IOException;
import java.io.InputStream;
import java.lang.foreign.Arena;
import java.lang.foreign.SymbolLookup;

import static dev.pfaff.altargui.glfw.GLGlobals.GLFW;
import static dev.pfaff.altargui.util.BasicFuture.executeOn;

final class HostImpl implements Host {
	private static final ConstantValue<Float> CONSTANT_VALUE_1F = new ConstantValue<>(1f);

	private static final IWindow WINDOW = new IWindow() {
		@Override
		public @NonNull Value<@NonNull Float> scaleFactorListenable() {
			return CONSTANT_VALUE_1F;
		}
	};

	private final Application application;

	public HostImpl(Application application) {
		this.application = application;
	}

	@Override
	public @NonNull GL gl() {
		return GLGlobals.GL;
	}

	@Override
	public @NonNull IWindow window() {
		return WINDOW;
	}

	@Override
	// this nullability warning is wrong
	public @Nullable String getClipboardString() {
		return executeOn(this.application.window::getClipboardString, this.application::submitToMainThreadOrRun);
	}

	@Override
	public void setClipboardString(@NonNull String string) {
		executeOn(() -> {
					this.application.window.setClipboardString(string);
					return null;
				},
				this.application::submitToMainThreadOrRun);
	}

	@Override
	public @NonNull SymbolLookup loadLibImage() {
		InputStream inputStream = HostImpl.class.getResourceAsStream(switch (OS.HOST) {
			case OS.Linux _ when Arch.HOST instanceof Arch.aarch64 -> "/natives/libimage_c-linux-arm64.so";
			case OS.macOS _ when Arch.HOST instanceof Arch.amd64 -> "/natives/libimage_c-macos-amd64.dylib";
			case OS.Windows _ when Arch.HOST instanceof Arch.amd64 -> "/natives/image_c-windows-amd64.dll";
			default -> throw dev.pfaff.glass.Host.makeUnsupportedOSArchException();
		});
		assert inputStream != null;
		try {
			return BundledLibraryLinker.createLookup(inputStream, Arena.ofAuto());
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public ContextForker contextForker() {
		return () -> {
			var window = BasicFuture.executeOn(() -> {
				return GLFWwindow.create(GLFW, 1, 1, "", null, this.application.window);
			}, GLThread::submitToGLThread);
			window.makeContextCurrent();
			return new ForkedContext() {
				@Override
				public GL gl() {
					return GLGlobals.GL;
				}

				@Override
				public void close() {
					window.destroy();
				}
			};
		};
	}
}
